/*
 * Copyright (C) 2008 Romain Guy
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.curiouscreature.android.shelves.provider;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import android.util.Xml;
import android.view.InflateException;
import android.net.Uri;

/**
 * Amazon implementation of {@link BooksStore}.
 */
class AmazonBooksStore extends BooksStore {
    // IMPORTANT: Replace this Amazon API key with your own
    private static final String API_KEY = "10P4HQT5BQEARZ1PWV82";

    private static final String API_REST_HOST = "ecs.amazonaws.com";
    private static final String API_REST_URL = "/onca/xml";

    private static final String API_ITEM_LOOKUP = "ItemLookup";
    private static final String API_ITEM_SEARCH = "ItemSearch";

    private static final String PARAM_API_KEY = "AWSAccessKeyId";
    private static final String PARAM_API_VERSION = "Version";
    private static final String PARAM_API_SERVICE = "Service";
    private static final String PARAM_OPERATION = "Operation";
    private static final String PARAM_IDTYPE = "IdType";
    private static final String PARAM_SEARCHINDEX = "SearchIndex";
    private static final String PARAM_ITEMID = "ItemId";
    private static final String PARAM_RESPONSEGROUP = "ResponseGroup";
    private static final String PARAM_KEYWORDS = "Keywords";

    private static final String VALUE_SERVICE = "AWSECommerceService";
    private static final String VALUE_VERSION = "2008-08-19";
    private static final String VALUE_IDTYPE_ISBN = "ISBN";
    private static final String VALUE_SEARCHINDEX_BOOKS = "Books";
    private static final String VALUE_RESPONSEGROUP_MEDIUM = "Medium";

    private static final String RESPONSE_TAG_REQUEST = "Request";
    private static final String RESPONSE_TAG_ISVALID = "IsValid";
    private static final String RESPONSE_TAG_ERRORS = "Errors";
    private static final String RESPONSE_TAG_ASIN = "ASIN";
    private static final String RESPONSE_TAG_DETAILPAGEURL = "DetailPageURL";
    private static final String RESPONSE_TAG_ITEMATTRIBUTES = "ItemAttributes";
    private static final String RESPONSE_TAG_AUTHOR = "Author";
    private static final String RESPONSE_TAG_LABEL = "Label";
    private static final String RESPONSE_TAG_ISBN = "ISBN";
    private static final String RESPONSE_TAG_EAN = "EAN";
    private static final String RESPONSE_TAG_NUMBEROFPAGES = "NumberOfPages";
    private static final String RESPONSE_TAG_PUBLICATIONDATE = "PublicationDate";
    private static final String RESPONSE_TAG_TITLE = "Title";
    private static final String RESPONSE_TAG_EDITORIALREVIEW = "EditorialReview";
    private static final String RESPONSE_TAG_SOURCE = "Source";
    private static final String RESPONSE_TAG_CONTENT = "Content";
    private static final String RESPONSE_TAG_IMAGESET = "ImageSet";
    // private static final String RESPONSE_TAG_SWATCHIMAGE = "SwatchImage";
    // private static final String RESPONSE_TAG_SMALLIMAGE = "SmallImage";
    private static final String RESPONSE_TAG_THUMBNAILIMAGE = "ThumbnailImage";
    private static final String RESPONSE_TAG_TINYIMAGE = "TinyImage";
    // private static final String RESPONSE_TAG_MEDIUMIMAGE = "MediumImage";
    // private static final String RESPONSE_TAG_LARGEIMAGE = "LargeImage";
    private static final String RESPONSE_TAG_URL = "URL";
    private static final String RESPONSE_TAG_ITEM = "Item";

    private static final String RESPONSE_ATTR_CATEGORY = "Category";
    private static final String RESPONSE_VALUE_CATEGORY_PRIMARY = "primary";

    AmazonBooksStore(String name, String label) {
        super(name, label, API_REST_HOST);
    }

    @Override
    Uri.Builder buildSearchBooksQuery(String query) {
        final Uri.Builder uri = buildGetMethod(API_ITEM_SEARCH);
        uri.appendQueryParameter(PARAM_KEYWORDS, query);
        uri.appendQueryParameter(PARAM_SEARCHINDEX, VALUE_SEARCHINDEX_BOOKS);
        uri.appendQueryParameter(PARAM_RESPONSEGROUP, VALUE_RESPONSEGROUP_MEDIUM);
        android.util.Log.d(LOG_TAG, uri.toString());
        return uri;
    }

    @Override
    Uri.Builder buildFindBookQuery(String id) {
        final Uri.Builder uri = buildGetMethod(API_ITEM_LOOKUP);
        uri.appendQueryParameter(PARAM_IDTYPE, VALUE_IDTYPE_ISBN);
        uri.appendQueryParameter(PARAM_SEARCHINDEX, VALUE_SEARCHINDEX_BOOKS);
        uri.appendQueryParameter(PARAM_RESPONSEGROUP, VALUE_RESPONSEGROUP_MEDIUM);
        uri.appendQueryParameter(PARAM_ITEMID, id);
        return uri;
    }

    @Override
    boolean parseBook(XmlPullParser parser, Book book) throws XmlPullParserException,
            IOException {

        int type;
        String name;
        final int depth = parser.getDepth();

        while (((type = parser.next()) != XmlPullParser.END_TAG ||
                parser.getDepth() > depth) && type != XmlPullParser.END_DOCUMENT) {
            if (type != XmlPullParser.START_TAG) {
                continue;
            }

            name = parser.getName();

            if (RESPONSE_TAG_ASIN.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mInternalId = parser.getText();
                }
            } else if (RESPONSE_TAG_DETAILPAGEURL.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mDetailsUrl = parser.getText();
                }
            } else if (RESPONSE_TAG_IMAGESET.equals(name)) {
                if (RESPONSE_VALUE_CATEGORY_PRIMARY.equals(
                        parser.getAttributeValue(null, RESPONSE_ATTR_CATEGORY))) {
                    parseImageSet(parser, book);
                }
            } else if (RESPONSE_TAG_ITEMATTRIBUTES.equals(name)) {
                parseItemAttributes(parser, book);
            } else if (RESPONSE_TAG_EDITORIALREVIEW.equals(name)) {
                parseEditorialReview(parser, book);
            } else if (RESPONSE_TAG_ERRORS.equals(name)) {
                return false;
            }
        }

        return true;
    }

    @Override
    boolean findNextBook(XmlPullParser parser) throws XmlPullParserException, IOException {
        if (RESPONSE_TAG_ITEM.equals(parser.getName())) {
            return true;
        }

        int type;
        final int depth = parser.getDepth();

        while (((type = parser.next()) != XmlPullParser.END_TAG ||
                parser.getDepth() > depth) && type != XmlPullParser.END_DOCUMENT) {
            if (type != XmlPullParser.START_TAG) {
                continue;
            }

            if (RESPONSE_TAG_ITEM.equals(parser.getName())) {
                return true;
            }
        }

        return false;
    }

    private void parseEditorialReview(XmlPullParser parser, Book book) throws IOException,
            XmlPullParserException {

        int type;
        String name;
        String source = null;
        String content = null;
        final int depth = parser.getDepth();

        while (((type = parser.next()) != XmlPullParser.END_TAG ||
                parser.getDepth() > depth) && type != XmlPullParser.END_DOCUMENT) {
            if (type != XmlPullParser.START_TAG) {
                continue;
            }

            name = parser.getName();
            if (RESPONSE_TAG_SOURCE.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    source = parser.getText();
                }
            } else if (RESPONSE_TAG_CONTENT.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    content = parser.getText();
                }
            }
        }

        if (content != null) {
            book.mDescriptions.add(new Description(source, content));
        }
    }

    private void parseItemAttributes(XmlPullParser parser, Book book) throws IOException,
            XmlPullParserException {

        int type;
        String name;
        final int depth = parser.getDepth();

        while (((type = parser.next()) != XmlPullParser.END_TAG ||
                parser.getDepth() > depth) && type != XmlPullParser.END_DOCUMENT) {
            if (type != XmlPullParser.START_TAG) {
                continue;
            }

            name = parser.getName();
            if (RESPONSE_TAG_AUTHOR.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mAuthors.add(parser.getText());
                }
            } else if (RESPONSE_TAG_EAN.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mEan = parser.getText();
                }
            } else if (RESPONSE_TAG_ISBN.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mIsbn = parser.getText();
                }
            } else if (RESPONSE_TAG_NUMBEROFPAGES.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mPages = Integer.parseInt(parser.getText());
                }
            } else if (RESPONSE_TAG_PUBLICATIONDATE.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    final SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd");
                    try {
                        book.mPublicationDate = format.parse(parser.getText());
                    } catch (ParseException e) {
                        // Ignore
                    }
                }
            } else if (RESPONSE_TAG_TITLE.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mTitle = parser.getText();
                }
            } else if (RESPONSE_TAG_LABEL.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mPublisher = parser.getText();
                }
            }
        }
    }

    private void parseImageSet(XmlPullParser parser, Book book) throws IOException,
            XmlPullParserException {

        int type;
        String name;
        final int depth = parser.getDepth();

        while (((type = parser.next()) != XmlPullParser.END_TAG ||
                parser.getDepth() > depth) && type != XmlPullParser.END_DOCUMENT) {
            if (type != XmlPullParser.START_TAG) {
                continue;
            }

            name = parser.getName();
            if (RESPONSE_TAG_TINYIMAGE.equals(name)) {
                parseImage(parser, ImageSize.TINY, book);
            } else if (RESPONSE_TAG_THUMBNAILIMAGE.equals(name)) {
                parseImage(parser, ImageSize.THUMBNAIL, book);
            }

            // if (RESPONSE_TAG_SWATCHIMAGE.equals(name)) {
            //     parseImage(parser, ImageSize.SWATCH, book);
            // } else if (RESPONSE_TAG_SMALLIMAGE.equals(name)) {
            //     parseImage(parser, ImageSize.SMALL, book);
            // } else if (RESPONSE_TAG_THUMBNAILIMAGE.equals(name)) {
            //     parseImage(parser, ImageSize.THUMBNAIL, book);
            // } else if (RESPONSE_TAG_TINYIMAGE.equals(name)) {
            //     parseImage(parser, ImageSize.TINY, book);
            // } else if (RESPONSE_TAG_MEDIUMIMAGE.equals(name)) {
            //     parseImage(parser, ImageSize.MEDIUM, book);
            // } else if (RESPONSE_TAG_LARGEIMAGE.equals(name)) {
            //     parseImage(parser, ImageSize.LARGE, book);
            // }
        }
    }

    private void parseImage(XmlPullParser parser, ImageSize size, Book book) throws IOException,
            XmlPullParserException {

        int type;
        String name;
        final int depth = parser.getDepth();

        while (((type = parser.next()) != XmlPullParser.END_TAG ||
                parser.getDepth() > depth) && type != XmlPullParser.END_DOCUMENT) {
            if (type != XmlPullParser.START_TAG) {
                continue;
            }

            name = parser.getName();
            if (RESPONSE_TAG_URL.equals(name)) {
                if (parser.next() == XmlPullParser.TEXT) {
                    book.mImages.put(size, parser.getText());
                }
            }
        }
    }

    @Override
    void parseResponse(InputStream in, ResponseParser responseParser) throws IOException {
        final XmlPullParser parser = Xml.newPullParser();
        try {
            parser.setInput(new InputStreamReader(in));

            int type;
            while ((type = parser.next()) != XmlPullParser.START_TAG &&
                    type != XmlPullParser.END_DOCUMENT) {
                // Empty
            }

            if (type != XmlPullParser.START_TAG) {
                throw new InflateException(parser.getPositionDescription()
                        + ": No start tag found!");
            }

            String name;
            boolean valid = false;
            final int topDepth = parser.getDepth();

            while (((type = parser.next()) != XmlPullParser.END_TAG ||
                    parser.getDepth() > topDepth) && type != XmlPullParser.END_DOCUMENT) {

                if (type != XmlPullParser.START_TAG) {
                    continue;
                }

                name = parser.getName();

                if (RESPONSE_TAG_REQUEST.equals(name)) {
                    valid = isRequestValid(parser);
                    break;
                }
            }

            if (valid) responseParser.parseResponse(parser);

        } catch (XmlPullParserException e) {
            final IOException ioe = new IOException("Could not parse the response");
            ioe.initCause(e);
            throw ioe;
        }
    }

    private boolean isRequestValid(XmlPullParser parser) throws XmlPullParserException,
            IOException {

        int type;
        String name;
        boolean valid = false;
        final int depth = parser.getDepth();

        while (((type = parser.next()) != XmlPullParser.END_TAG ||
                parser.getDepth() > depth) && type != XmlPullParser.END_DOCUMENT) {

            if (type != XmlPullParser.START_TAG) {
                continue;
            }

            name = parser.getName();

            if (RESPONSE_TAG_ISVALID.equals(name)) {
                if (parser.next() != XmlPullParser.TEXT ||
                        !"True".equals(parser.getText())) {
                    throw new IOException("Invalid request");
                } else {
                    valid = true;
                }
            } else if (RESPONSE_TAG_ERRORS.equals(name)) {
                valid = false;
            }
        }

        return valid;
    }

    /**
     * Builds an HTTP GET request for the specified API method. The returned request
     * contains the web service path, the query parameter for the API KEY and the query
     * parameter for the specified method.
     *
     * @param method The API method to invoke.
     *
     * @return A Uri.Builder containing the GET path, the API key and the method already
     *         encoded.
     */
    private static Uri.Builder buildGetMethod(String method) {
        final Uri.Builder builder = new Uri.Builder();
        builder.path(API_REST_URL);
        builder.appendQueryParameter(PARAM_API_KEY, API_KEY);
        builder.appendQueryParameter(PARAM_API_VERSION, VALUE_VERSION);
        builder.appendQueryParameter(PARAM_API_SERVICE, VALUE_SERVICE);
        builder.appendQueryParameter(PARAM_OPERATION, method);
        return builder;
    }
}
